# python3-opsi
# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
# and published under the Terms of the GNU Affero General Public License.
# Text of the AGPL: http://www.gnu.org/licenses/agpl-3.0-standalone.html
# credits: http://www.opsi.org/credits/

#***************************************************************************
# Subversion:
# $Revision$
# $Author$
# $Date$
#***************************************************************************

#

try:
    import ConfigParser
except:
    import configparser
from OPSI.Util.File import IniFile
from OPSI.Util.File import TextFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from OPSI.Util import blowfishDecrypt
from hashlib import md5
try:
    from twisted.conch.ssh import keys
except:
    pass
try:
    from urlparse import urlparse
except:
    from urllib.parse import urlparse
from os.path import splitext, basename
import platform
import re, xml.dom.minidom, os, gzip
import crypt
from base64 import b64encode
from os import urandom

if not bootimageVersion or (int(bootimageVersion) < 20210618):
    raise Exception(u"This product requires opsi-linux-bootimage version >= 20210618")
try:
    if not backend.backend_info()['opsiVersion'].startswith('4.'):
        raise ValueError('No opsi 4.x backend present')
except Exception:
    raise Exception(u"This is a opsi 4.x product")

isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

if isOpsi43:
    show_message(u"Installing Product: %s" % (productId), 10)

opsiVersion = backend.backend_info()['opsiVersion']

# Do hardware inventory
logger.notice(u"Fetching opsi hw audit configuration")
hwconfig = backend.auditHardware_getConfig()
if isOpsi43:
    show_message(u"Running hardware inventory", 10)
logger.notice(u"Running hardware inventory")
auditHardwareOnHosts = auditHardware(config = hwconfig, hostId = clientId)

logger.notice(u"Sending hardware information to service")
backend.auditHardwareOnHost_setObsolete(clientId)
backend.auditHardwareOnHost_createObjects(auditHardwareOnHosts)

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')
exec(open('/tmp/opsisetuplib.py').read())

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsipreparelib.py','/tmp')
exec(open('/tmp/opsipreparelib.py').read())

initOpsiSetupLib()

useGpt = False
inUefi = inUefiMode()
inSecureboot = False
if inUefi:
    try:
        inSecureboot = isSecureboot()
    except NameError:
        isSecureboot = False

    if inSecureboot:
        logger.notice(u"we are running in SecureBoot mode")
        scriptMessageSubject.setMessage(u"we are running in SecureBoot mode")
    else:
        logger.notice(u"we are not running in SecureBoot mode")
        scriptMessageSubject.setMessage(u"we are not running in SecureBoot mode")
    useGpt = True

if useGpt:
    logger.notice(u"Will use GPT")
else:
    logger.notice(u"Will use MBR Partitions")

# create network path
if (depot._url.startswith('smb:') or depot._url.startswith('cifs:')):
    match = re.search('^smb://([^/]+)/([^/]+)(.*)$', depot._url, re.IGNORECASE)
    if not match:
        raise Exception("Bad depot-URL '%s'" % depot._url)
else:
    match = re.search('^webdavs://([^/]+):([^/]+)(.*)$', depot._url, re.IGNORECASE)
    if not match:
        raise Exception("Bad depot-URL '%s'" % depot._url)

depotDir = '/var/lib/opsi/depot'

proxy = productProperties.get('proxy','')

depotAddress = depotUrl.split('/')[2].replace(':4447','')
install_opsi_client_agent = False
do_opsi_client = "false"
packages_to_install_for_opsi_client_agent = ''
opsi_client_agent_installation = ''
epel_repo_installation = ''
if (str(productProperties.get('install_opsi-client-agent', 'true')).lower() == 'true'):
    install_opsi_client_agent = True
    do_opsi_client = "true"
    if (productId == 'alma8') or (productId == 'rocky8') or (productId == 'alma9') or (productId == 'rocky9'):
        epel_repo_installation = '%post --interpreter=/bin/bash --log=/var/log/ks-post-epel-log.log\n\
set -x\n\
dnf -y install epel-release\n\
%end'
    if productId == 'centos70':
        epel_repo_installation = '%post --interpreter=/bin/bash --log=/var/log/ks-post-epel-log.log\n\
yum -y install epel-release\n\
%end'
    if productId == 'centos8':
        epel_repo_installation = '%post --interpreter=/bin/bash --log=/var/log/ks-post-epel-log.log\n\
dnf -y install epel-release epel-next-release\n\
%end'
    if productId == 'redhat70':
        epel_repo_installation = '%post --interpreter=/bin/bash --log=/var/log/ks-post-epel-log.log\n\
yum -y install https://dl.fedoraproject.org/pub/epel/epel-release-latest-7.noarch.rpm\n\
%end'
    if productId == 'redhat8':
        epel_repo_installation = '%post --interpreter=/bin/bash --log=/var/log/ks-post-epel-log.log\n\
dnf -y install https://dl.fedoraproject.org/pub/epel/epel-release-latest-8.noarch.rpm\n\
%end'
    if productId == 'redhat9':
        epel_repo_installation = '%post --interpreter=/bin/bash --log=/var/log/ks-post-epel-log.log\n\
dnf -y install https://dl.fedoraproject.org/pub/epel/epel-release-latest-9.noarch.rpm\n\
%end'
    packages_to_install_for_opsi_client_agent = 'redhat-lsb-core cifs-utils xterm bind-utils wget nmap-ncat'
    opsi_client_agent_installation = '%%post --interpreter=/bin/bash --log=/var/log/ks-post-olca-log.log\n\
set -x\n\
yum -y install %s\n\
mkdir -p /var/tmp/opsi-linux-client-agent\n\
wget -r -nH --no-check-certificate --cut-dirs=2 --no-parent --http-user #@client_id*# --http-passwd #@client_key*# --reject="index.html*" https://#@depotAddress*#:4447/depot/opsi-linux-client-agent/ -P /var/tmp/opsi-linux-client-agent\n\
chmod -R 777 /var/tmp/opsi-linux-client-agent\n\
echo "CLIENT_ID=#@client_id*#" >> /var/tmp/opsi-linux-client-agent/files/postinst.d/postinst.env\n\
echo "CLIENT_KEY=#@client_key*#" >> /var/tmp/opsi-linux-client-agent/files/postinst.d/postinst.env\n\
echo "SERVICE_ADDRESS=https://#@depotAddress*#:4447" >> /var/tmp/opsi-linux-client-agent/files/postinst.d/postinst.env\n\
wget -r -nH --no-check-certificate --cut-dirs=3 --no-parent --http-user #@client_id*# --http-passwd #@client_key*# --reject="index.html*" https://#@depotAddress*#:4447/depot/#@productid*#/opsi/postinst.sh -P /var/tmp/opsi-linux-client-agent/files/\n\
chmod +x /var/tmp/opsi-linux-client-agent/files/postinst.sh\n\
wget -r -nH --no-check-certificate --cut-dirs=2 --no-parent --http-user #@client_id*# --http-passwd #@client_key*# --reject="index.html*" https://#@depotAddress*#:4447/depot/#@productid*#/olca-install.service -P /etc/systemd/system/\n\
sed -i s/service_ip_address/#@depotAddress*#/g /etc/systemd/system/olca-install.service\n\
systemctl enable olca-install.service\n\
%%end' % (packages_to_install_for_opsi_client_agent)
    packages_to_install_for_opsi_client_agent = packages_to_install_for_opsi_client_agent.replace(" ", "\\n")


installation_url = productProperties.get('installation_url', '')
installation_source = productProperties.get('installation_source', 'url')
installation_source_url = ''
if (installation_source == 'isocontent'):
    installation_url = f"https://{fqdn}:{pckey}@{depotUrl.split('/')[2].replace(':4447','')}:4447/depot/{productId}/isocontent/"
    logger.notice(f"installation_url = {installation_url}")
    preskript = ''
    preskript = f'%pre --log /tmp/ca-cert.log\n\
wget --no-check-certificate -P /etc/pki/ca-trust/source/anchors/ https://{fqdn}:{pckey}@{depotAddress}:4447/ssl/opsi-ca-cert.pem\n\
update-ca-trust\n\
%end\n'


install_unattended = True
if (str(productProperties.get('install_unattended', 'true')).lower() == 'false'):
    install_unattended = False

installation_method = productProperties.get('installation_method', 'kexec')
if inSecureboot:
    installation_method = 'reboot'
    logger.notice("We are running in SecureBoot mode, enforcing reboot installation_method")

locale = productProperties.get('locale','de_DE')
language = locale.split('_')[0]
country = locale.split('_')[1]
console_keymap = productProperties.get('console_keymap', 'de')

partition_disk = productProperties.get('partition_disk', 'first')
partition_method = productProperties.get('partition_method', 'lvm')
filesystem_type = productProperties.get('filesystem_type', 'xfs')

if (productId == 'centos70') or (productId == 'redhat70'):
    # can't use btrfs as fstype in the kickstart file
    if (filesystem_type == 'btrfs'):
        filesystem_type = ''

partition_file = productProperties.get('partition_file', 'auto')
if partition_file == 'auto':
    logger.notice(u"Default partitioning is used")
    partition_scheme = ''
    partition_method_custom = ''
else:
    logger.notice(u"Custom partition scheme (%s) found and will be used" % partition_file)
    depot.copy(SCRIPT_PATH+'/custom/%s' % partition_file,'/tmp')
    with open('/tmp/%s' % partition_file,'r') as myfile:
        partition_scheme = myfile.read()
    logger.notice(u"partition_scheme: %s" % (partition_scheme))
    partition_scheme = partition_scheme.replace("\n", "\\n")
    logger.notice(u"partition_scheme: %s" % (partition_scheme))
    partition_method_custom = '#'

my_broadcast = getBroadcastAddress(ipaddress, netmask)
my_network = getNetworkAddress(ipaddress, netmask)


if (productId.startswith('redhat')):
    registration_user=""
    registration_pass=""
    subscription_register = False
    if (str(productProperties.get('subscription_register', 'true')).lower() == 'true'):
        subscription_register = True
    logger.notice(u"subscription_register:%s" % subscription_register)
    if subscription_register:
        logger.notice(u"softwareLicenseKey:%s" % softwareLicenseKey)
        if (softwareLicenseKey == ''):
            raise "No Licensekey found"
        else:
            logger.notice(u"splitting softwareLicenseKey for subscription_register")
            registration_user=softwareLicenseKey.split(':')[0]
            registration_pass=softwareLicenseKey.split(':')[1]
    if subscription_register:
        do_subscription_registration = ''
    else:
        do_subscription_registration = '#'

# Get and encrypt passwords
# user
username = productProperties.get('user_name', 'user')
userpassclear = productProperties.get('user_password', 'linux123')
userpassenc = shadow_password_hash(userpassclear)

if (username == ''):
    create_user = '#'
else:
    create_user = ''

# root
rootpassclear = productProperties.get('root_password', 'linux123')
rootpassenc = shadow_password_hash(rootpassclear)

selinux_mode = productProperties.get('selinux_mode', 'permissive')
partition_method = productProperties.get('partition_method', 'lvm')

if (proxy == ''):
    do_proxy = 'false'
    proxy_addr = ''
else:
    do_proxy = 'true'
    proxy_addr = proxy

opsi_client_agent_install_script = ''


kernel_boot_options = productProperties.get('kernel_boot_options', '')
kernel_boot_options = kernel_boot_options.replace(",", " ")
kernel_boot_options_installer = productProperties.get('kernel_boot_options_installer', '')
kernel_boot_options_installer = kernel_boot_options_installer.replace(",", " ")
packages_to_install = productProperties.get('packages_to_install', '')
packages_to_install = packages_to_install.replace(",", "\\n")

# get machine architecture
myarchitecture = getArchitecture()
if myarchitecture == 'amd64':
    scriptMessageSubject.setMessage(u"Installing 64 Bit system....")
    logger.notice(u"Installing 64 Bit system....")
    myarchitectureBits = '64'
else:
    scriptMessageSubject.setMessage("Installing 32 Bit system....")
    logger.notice(u"Installing 32 Bit system....")
    myarchitectureBits = '32'

if isOpsi43:
    show_message("Preparing disk", 20)
scriptMessageSubject.setMessage("Create temporary place on disk.... ")
logger.notice(u"Create temporary place on disk.... ")
#prepareDisk()
target = '/mnt/hd'
systemPartitionNumber = 1
if inUefi:
    systemPartitionNumber = 3
    if installation_method == 'reboot':
        grubPartitionNumber = 3
        systemPartitionNumber +=1
        grubPartitionSizeM = 2000
        grubPartitionLabel = 'INSTALLER'
systemPartitionSizeM = 2000
systemPartitionLabel  = 'LINUX'
blockAlignment = True

logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
lvmClearVolumeGroups()
execute('sfdisk -l')

# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
# use first disk
disk = disks[0]
if partition_disk == "first":
    partition_disk = disk.device

if useGpt:
    logger.notice(u"Clear partition table of %s" % (disk.device))
    execute('/usr/sbin/sgdisk --zap-all %s' % (disk.device))
    execute('/usr/sbin/sgdisk --clear -g -p %s' % (disk.device))
    if inUefi:
        currentbootentry = uefiGetActBootEntry()
        if currentbootentry:
            uefiNetbootLabel = currentbootentry['label']
            logger.notice('Current (net) boot entry: %s' % uefiNetbootLabel)
            backend.config_createUnicode("clientConfig.uefiNetbootLabel","clientspecific uefi firmware netboot entry label") 
            backend.configState_create("clientConfig.uefiNetbootLabel", clientId, [ uefiNetbootLabel ] )
        else:
            logger.notice('No usable current (net) boot entry found')
        deleteAllEntriesFromEFIBootmgr('opsitempwinpe')
        deleteAllEntriesFromEFIBootmgr('Windows Boot Manager')
        deleteAllEntriesFromEFIBootmgr('ubuntu')
        deleteAllEntriesFromEFIBootmgr('debian')
        deleteAllEntriesFromEFIBootmgr('installer')
        deleteAllEntriesFromEFIBootmgr('opsibootimage')
        deleteAllEntriesFromEFIBootmgr('CentOS')
        deleteAllEntriesFromEFIBootmgr('Red Hat')
        deleteAllEntriesFromEFIBootmgr('Oracle')
        logger.notice(u"Create EFI system partition on %s1" % (disk.device))
        execute('/usr/sbin/sgdisk -g --new 1::100M -t 1:ef00 -c 1:\"EFI system partition\" -p %s' % (disk.device))
        execute('/bin/sleep 2')
        execute('/usr/sbin/sgdisk -p %s' % (disk.device))
        logger.notice(u"Make file system on EFI system partition on %s1" % (disk.device))
        normalDiskDevices = ['/dev/sda', '/dev/vda', '/dev/sdb']
        if any(normalDisk in disk.device for normalDisk in normalDiskDevices):
            execute('/sbin/mkfs.vfat %s1' % (disk.device))
        else:
            execute('/sbin/mkfs.vfat %sp1' % (disk.device))
        logger.notice(u"Create MS reserved partition on %s2" % (disk.device))
        execute('/usr/sbin/sgdisk -g --new 2::+128M -t 2:0c01 -c 2:\"MS reserved\" -p %s' % (disk.device))
else:
    disk.deletePartitionTable()

if installation_method == 'reboot':
    if inUefi:
        currentbootentry = uefiGetActBootEntry()
        if currentbootentry:
            uefiNetbootLabel = currentbootentry['label']
            logger.notice('Current (net) boot entry: %s' % uefiNetbootLabel)
            backend.config_createUnicode("clientConfig.uefiNetbootLabel","clientspecific uefi firmware netboot entry label") 
            backend.configState_create("clientConfig.uefiNetbootLabel", clientId, [ uefiNetbootLabel ] )
        else:
            logger.notice('No usable current (net) boot entry found')
        grubTarget = '/mnt/grub/'
        checkCreateDir(grubTarget)
        execute('/usr/sbin/sgdisk --info=%d %s' % (grubPartitionNumber, disk.device))
        createPartitionEx("installer", grubPartitionNumber, grubPartitionLabel, 0, grubPartitionSizeM, gptAutoAppend = True, boot = True, gptAttribute = "63" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "fat32", gptFileSystem="0700")
        execute('/usr/sbin/sgdisk -t %d:0700 --attributes %d:set:63 --attributes %d:clear:0 -p %s' % (grubPartitionNumber, grubPartitionNumber, grubPartitionNumber, disk.device))
        testPartitionEx(disk.device, grubPartitionNumber, useGpt)
        checkClearPartition("none", disk.device, grubPartitionNumber)
        createFilesystemEx("bootimage", grubPartitionNumber, "bootimage", "fat32", disk.device , useGpt)
        checkMount(grubTarget, disk.device, grubPartitionNumber, "vfat")
        logger.notice("Copying files to partition %d" % grubPartitionNumber)
        depot.copy('/%s/installfiles64/*' % productId, grubTarget)

        customisationLine = 'inst.keymap=%s %s %s' % (console_keymap, kernel_boot_options, kernel_boot_options_installer)
        newLines = []
        with open('%s/grub/grub.cfg' % (grubTarget), 'r+') as grubConfig:
            for line in grubConfig:
                if line.strip().startswith('linux'):
                    newLines.append('%s %s\n' % (line.rstrip(), customisationLine.rstrip()))
                    continue
                newLines.append(line)

        with open('%s/grub/grub.cfg' % (grubTarget), 'w') as grubConfig:
            grubConfig.writelines(newLines)

    else:
        logger.warning(u"reboot installation method only works using UEFI. Reverting back to kexec method")
        installation_method = "kexec"

fdiskexitcode = os.system('%s %s >> /dev/null 2>&1' % (which('partprobe'), disk.device))
createPartitionEx("system", systemPartitionNumber, systemPartitionLabel, 0, systemPartitionSizeM, gptAutoAppend = True, boot = False, gptAttribute = "" , useGpt = useGpt, diskDevice = disk.device, fileSystem = "ext4", gptFileSystem="8300")
createFilesystemEx("system", systemPartitionNumber, systemPartitionLabel, "ext4", disk.device, useGpt)
systemPartitionDevicePath = getPartitionDevicePath(disk.device, systemPartitionNumber)

# Partition table deleted => delete all netboot product states
deletePocs = []
for poc in backend.productOnClient_getObjects(clientId = clientId):
    if ((poc.productType == 'NetbootProduct') and (poc.productId != productId)) or \
       ((poc.productType == 'LocalbootProduct') and (poc.installationStatus == 'not_installed') and (poc.actionRequest == 'none')):
        deletePocs.append(poc)
backend.productOnClient_deleteObjects(deletePocs)

if isOpsi43:
    show_message("Mounting partitions and copying data", 40)
# Mount temporary partition
checkCreateDir(target)
checkClearMountpoint(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkClearPartition(target, disk.device, systemPartitionNumber, systemPartitionDevicePath)
checkMount(target, disk.device, systemPartitionNumber, "ext4", systemPartitionDevicePath)
checkCreateDir(target+'/tmp')


# We use the package kernel and initrd, because only these are signed by us
scriptMessageSubject.setMessage("Getting kernel and initrd.... ")
logger.notice(u"Getting kernel and initrd.... ")
if (installation_source == 'isocontent'):
    try:
        depot.copy(SCRIPT_PATH + '/isocontent/isolinux/vmlinuz', target+'/tmp')
        depot.copy(SCRIPT_PATH + '/isocontent/isolinux/initrd.img', target+'/tmp')
    except Exception as fnferror:
        raise Exception("isocontent directory doesn't seem to be filled")
else:
    execute(f'wget {installation_url}/isolinux/vmlinuz -P {target}/tmp/')
    execute(f'wget {installation_url}/isolinux/initrd.img -P {target}/tmp/')

scriptMessageSubject.setMessage("Uncompressing initrd.... ")
logger.notice(u"Uncompressing initrd.... ")
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
try:
    execute("file %s/tmp/initrd*" % (target))
except:
    pass
execute("mv %s/tmp/initrd.img %s/tmp/initrd.xz" % (target,target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("xz -d %s/tmp/initrd.xz"% (target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass

scriptMessageSubject.setMessage("Patching initrd.... ")
logger.notice(u"Mounting initrd ....")
execute("mkdir -p %s/tmp/mnt" % (target))
execute("cd %s/tmp/mnt ; cpio -idm -H newc < ../initrd" % (target))

logger.notice(u"Copy profile for patch....")
kickstart = productProperties.get('kickstart', 'auto')
if kickstart == 'auto':
    logger.notice(u"Default kickstart is used")
    depot.copy(SCRIPT_PATH+'/%s.cfg' % productId, target+'/tmp/installation_profile')
else:
    logger.notice(u"Custom installation profile (%s) found and will be used" % kickstart)
    depot.copy(SCRIPT_PATH+'/custom/%s' % kickstart, target+'/tmp/installation_profile')

logger.notice(u"Prepare patch file....")
opsipatchscript='/sbin/opsipatchscript.sh'
with open('%s/tmp/mnt%s' % (target,opsipatchscript), "a") as myfile:
    myfile.write('#! /bin/sh\n')
    myfile.write('set -x\n')
    myfile.write('chroot /target install-keymap %s\n' % console_keymap)
    myfile.write('cp /opsitmp/keyboard /target/etc/default/keyboard\n')
execute('chmod u+x %s/tmp/mnt%s\n' % (target,opsipatchscript))
logger.notice('Here comes the %s/tmp/mnt%s' % (target,opsipatchscript))
execute('cat %s/tmp/mnt%s' % (target,opsipatchscript))

if isOpsi43:
    show_message("patching files", 90)
logger.notice(u"Prepare patch input....")
with open(PATCHA_IN, "a") as myfile:
    myfile.write('networkaddress=%s\n' % my_network)
    myfile.write('broadcastaddress=%s\n' % my_broadcast)
    myfile.write('nameserveraddress=%s\n' % nameserver)
    myfile.write('do_opsi_client=%s\n' % do_opsi_client)
    myfile.write('username=%s\n' % username)
    myfile.write('create_user=%s\n' % create_user)
    myfile.write('userpassenc=%s\n' % userpassenc)
    myfile.write('rootpassenc=%s\n' % rootpassenc)
    myfile.write('do_proxy=%s\n' % do_proxy)
    myfile.write('proxy_addr=%s\n' % proxy_addr)
    myfile.write('selinux_mode=%s\n' % selinux_mode)
    myfile.write('kernel_boot_options=%s\n' % kernel_boot_options)
    myfile.write('kernel_boot_options_installer=%s\n' % kernel_boot_options_installer)
    myfile.write('packages_to_install=%s\n' % packages_to_install)
    myfile.write('packages_to_install_for_opsi_client_agent=%s\n' % packages_to_install_for_opsi_client_agent)
    myfile.write('installation_source_url=%s\n' % installation_source_url)
    myfile.write('installation_url=%s\n' % installation_url)
    myfile.write('partition_method=%s\n' % partition_method)
    myfile.write('partition_disk=%s\n' % partition_disk)
    myfile.write('partition_scheme=%s\n' % partition_scheme)
    myfile.write('partition_method_custom=%s\n' % partition_method_custom)
    myfile.write('filesystem_type=%s\n' % filesystem_type)
    #myfile.write('initrd_iso_filepath=%s\n' % initrd_iso_filepath)
    myfile.write('language=%s\n' % language)
    myfile.write('country=%s\n' % country)
    myfile.write('console_keymap=%s\n' % console_keymap)
    #myfile.write('opsistartupscript=%s\n' % opsistartupscript)
    #myfile.write('opsipatchscript=%s\n' % opsipatchscript)
    #myfile.write('repourl=%s\n' % repourl)
    myfile.write('service_ip_address=%s\n' % (nextserver.replace("/rpc", "").rstrip("/")))
    myfile.write('depotAddress=%s\n' % (depotUrl.split('/')[2].replace(':4447','')))
if (productId.startswith('redhat')):
    with open(PATCHA_IN, "a") as myfile:
        myfile.write('registration_user=%s\n' % registration_user)
        myfile.write('registration_pass=%s\n' % registration_pass)
        myfile.write('do_subscription_registration=%s\n' % do_subscription_registration)

logger.notice(u"Here comes the PATCHA_IN (%s):" % PATCHA_IN)
execute('cat %s' % (PATCHA_IN))

logger.notice(u"Patching installation_profile")
with open(target + '/tmp/installation_profile', 'r') as inputxml:
    filedata = inputxml.read()
    if install_opsi_client_agent == True:
        filedata = filedata.replace('#@opsi_client_agent_installation*#', opsi_client_agent_installation)
        filedata = filedata.replace('#@epel_repo_installation*#', epel_repo_installation)
    if (installation_source == 'isocontent'):
        filedata = filedata.replace('#@preskript*#', preskript)
    with open(target + '/tmp/installation_profile', 'w') as outputxml:
        outputxml.write(filedata)

execute('/usr/local/bin/patcha %s' % (target + '/tmp/installation_profile'))
execute('/usr/local/bin/patcha -f %s %s' % (PATCHA_IN, target + '/tmp/installation_profile'))

if install_unattended:
    logger.notice(u"Copy profile file to initrd...")
    copy(target+'/tmp/installation_profile', target+'/tmp/mnt/ks.cfg')
    logger.notice(u"Here comes the installation profile:")
    execute('cat %s/tmp/mnt/ks.cfg' % (target))

execute('mkdir -p %s/tmp/mnt/opsitmp/' % (target))
copy(PATCHA_IN, target+'/tmp/mnt/opsitmp/')

execute(f"wget --no-check-certificate -P /tmp/ https://{fqdn}:{pckey}@{depotAddress}:4447/ssl/opsi-ca-cert.pem")
execute(f"echo '' >> {target}/tmp/mnt/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem")
execute(f"echo '#opsi-server cert' >> {target}/tmp/mnt/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem")
execute(f"cat /tmp/opsi-ca-cert.pem >> {target}/tmp/mnt/etc/pki/ca-trust/extracted/pem/tls-ca-bundle.pem")

scriptMessageSubject.setMessage("Rebuild patched initrd.... ")
logger.notice(u"Rebuild patched initrd.... ")
execute("rm %s/tmp/initrd" % (target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass
execute("cd %s/tmp/mnt ; find . | cpio -o -H newc > %s/tmp/initrd" % (target, target))
try:
    execute("ls -l %s/tmp/initrd*" % (target))
except:
    pass

if installation_method == 'reboot':
    execute('/bin/efibootmgr -c --disk %s --part %s -L installer --loader \\\\BOOTX64.EFI' % (disk.device, grubPartitionNumber))
    setFirstUefiBoot('installer')
    setNextUefiBoot('installer')
    execute('/bin/efibootmgr -v')
    execute("cp %s/tmp/initrd* %s" % (target, grubTarget))
    execute("cp %s/tmp/vmlinuz %s" % (target, grubTarget))
    if (installation_source == 'isocontent'):
        depot.copy(f"{SCRIPT_PATH}/isocontent/EFI/BOOT/BOOTX64.EFI", grubTarget)
        depot.copy(f"{SCRIPT_PATH}/isocontent/EFI/BOOT/grubx64.efi", grubTarget)
    else:
        execute(f'wget {installation_url}/EFI/BOOT/BOOTX64.EFI -P {grubTarget}')
        execute(f'wget {installation_url}/EFI/BOOT/grubx64.efi -P {grubTarget}')
    execute(f"cp {grubTarget}/grub/grub.cfg {grubTarget}")
    execute("cp %s/tmp/mnt/ks.cfg %s" % (target, grubTarget))
else:
    scriptMessageSubject.setMessage("Prepare loading kernel.... ")
    logger.notice(u"Prepare loading kernel ....")
    kernel_file = target+'/tmp/vmlinuz'
    if (installation_source == 'isocontent'):
        append_line = f"inst.ks=file:/ks.cfg inst.repo=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/isocontent/ ip=dhcp ipv6.disable=1 inst.keymap={console_keymap} {kernel_boot_options} {kernel_boot_options_installer}"
    else:
        append_line = f"inst.ks=file:/ks.cfg inst.repo={installation_url} ip=dhcp ipv6.disable=1 inst.keymap={console_keymap} {kernel_boot_options} {kernel_boot_options_installer}"
        
    #append_line = f"inst.ks=file:/ks.cfg inst.repo=https://{fqdn}:{pckey}@{depotAddress}:4447/depot/{productId}/isocontent/  inst.keymap={console_keymap} {kernel_boot_options} {kernel_boot_options_installer}"
    inird_file = target+'/tmp/initrd'
    execute("%s --load %s  --append='%s' --initrd=%s" % (which('kexec'), kernel_file, append_line, inird_file))

logger.notice(u"handle_setupAfterInstall ....")
handle_setupAfterInstall()

# Set product installation status
productOnClient.setInstallationStatus('unknown')
#productOnClient.setActionProgress('installing')
productOnClient.setActionResult('')
productOnClient.setLastAction('setup')

#scriptMessageSubject.setMessage("Writing log to service.... ")
try:
    # Write log to service
    logger.notice(u"Writing log to service")

    log = codecs.open(LOG_FILE, 'r', 'utf-8', 'replace')
    data = log.read()
    log.close()

    logger.setFileLevel(LOG_ERROR)
    if backend:
        try:
            try:
                ui.getMessageBox().addText(_(u"Writing log to service\n"))
            except:
                pass
            backend.log_write(logType = 'bootimage', data = data, objectId = clientId, append = False)
        except Exception as e:
            logger.error(u"Failed to write log to service: %s" % e)

except Exception as e:
    logger.logException(e)

if installation_method == 'reboot':
    scriptMessageSubject.setMessage("Rebooting machine ")
    logger.notice(u"Rebooting machine")
    reboot()
else:
    scriptMessageSubject.setMessage("Execute new kernel.... ")
    logger.notice(u"Execute new kernel ....")
    execute("%s -e" % (which('kexec')))

